; MIT License

; Copyright (C) 2022 Z80Enjoyers / Luis Jesús Marhuenda Tendero (@ljmt1) / Alberto Vegara Belmonte (@avb52-ua) / Samuel Benjamín Oliva Bulpitt (@samuel0122)

; Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal 
; in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
; copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

; The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
; FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

;;======================================================================================================
.module Entity_Manager
;;======================================================================================================
.area _CODE
;;======================================================================================================
.include "cpctelera.h.s"
.include "man/entity.h.s"
.include "man/macros.h.s"
.include "cpct_functions.inc"
.globl cpct_memset_asm
;;======================================================================================================

;;======================================================================================================
;;  Private Values
;;
m_num_created_entities::
    .db 0

m_next_free_entity::
    .dw entities

entities::
    .ds max_entities * sizeof_e

m_zero_type_at_the_end:
    .dw #e_type_invalid




;======================================================================================================
;  Reserve space for a new entity
;  
;  Inputs:
;      -
;  Outputs:
;      - DE, pointer to the entity created
;      - IX, pointing to the next free entity space
;
man_entity_create::
    ld      de, (m_next_free_entity)

    ld      ix, (m_next_free_entity)
    ld      bc, #sizeof_e
    add     ix, bc
    ld      (m_next_free_entity), ix

    ld hl,  #m_num_created_entities
    inc (hl)
    
    ret

;======================================================================================================
;  Erases all entities
;  
;  Inputs:
;      -
;  Outputs:
;      -
;
man_entity_erase_all::

    ; (2B DE) array	Pointer to the first byte of the array to be filled up (starting point in memory)
    ; (1B A ) value	8-bit value to be set
    ; (2B BC) size	Number of bytes to be set (>= 2)
    ld  de,     #entities
    ld  bc,     #max_entities * sizeof_e
    ld  a,      #e_cmp_invalid
    call cpct_memset_asm    
    
    ; Contador a 0
    ld  hl,     #m_num_created_entities
    ld  (hl),   #0


    ; Next free entity al inicio
    ld  hl,     #entities
    ld  (m_next_free_entity),   hl
    
    ret


;======================================================================================================
;  Destroys an entity
;
;  Inputs:
;      - IX, Pointer to the entity to be destroyed
;
man_entity_destroy::
    push ix
    pop de
    
    ;; Obtengo el puntero al último entity de la lista. Esto es para sustituir el elemento a borrar por este último, evitando así huecos vacíos en el array.
    ld hl, (m_next_free_entity)
    ld bc, #-sizeof_e
    add hl, bc

    ;; Primero compruebo si se trata del mismo entity el último que el que quiero eliminar
    push hl     ; HL -> last entity
    sbc hl, de  ; DE -> * dead_e
    pop hl      ; Recupero HL

    ;; Si se trata del entity final, salto
    jr z, is_finalEntity    ; Si quiero eliminar el entity del final, simplemente quiero desactivarlo, con lo cual salto este paso.
        ;; Si no se trata del último, copio el último en *dead_e
        
        ;; En DE ya tengo el puntero al destino (*dead_e)
        ;; En HL ya tengo el puntero del origen (ultimo entity)
        ld bc, #sizeof_e     ; BC -> tamaño de bytes a copiar
        push hl
        push de
        call cpct_memcpy_asm    ; Destruye los registros HL y DE
        pop de
        pop hl

    is_finalEntity:

    ;; Ahora quiero deshabilitar el último entity, guardar como nex_free_entity este nuevo espacio librado, y disminuir el contador de entities
    ex de, hl
    ld hl, #e_cmps
    add hl, de

    ld (hl), #e_cmp_invalid     ;; Cambio el flag component para que no sea procesado

    ld hl, #e_type
    add hl, de

    ld (hl), #e_type_invalid    ;; Cambio el flag type a inválido

    ld (m_next_free_entity), de

    ld hl, #m_num_created_entities
    dec (hl)

    ret

;;======================================================================================================
;;  Destroys an entity
;;  Inputs:
;;      - IX, Pointer to the entity to be set to destroyed
;;
man_entity_set4destroy::
    ;; Activo el flag dead en el entity.
    ld      a, e_type(ix)       ; Cargo el type
    or      #e_type_dead        ; Opero con un OR al A
    ld      e_type(ix), a        ; Guardo el type en el entity
    
    ret

;======================================================================================================
;  Loop through the array of entities and call the funcion passed
;  as a pointer with the entites whos match the signature
;
;  Inputs:
;      - HL, pointer to the function to be called
;      - B, (signature) type of the entity to be called when match the type
;
;  Function's Inputs:
;      - IX, pointer to entity to update
;
man_entity_forall_matching::
    ld      ix, #entities-sizeof_e      ;; Point before the first entity in order to make the loop easier

    startWhile:

        NextEntity ix   ;; Point to the next entity

        IsValid ix, while_end
            ;; If the signature does not match, go to the next entity
            SignatureDoesNotMatch ix, b, startWhile

            ;; If the signature matches, call function pointer
            SaveRegisters
            ld      de, #returnPoint
            push    de

            jp      (hl)    ;; Jump to the function pointer

            returnPoint:
            RecoverRegisters

            jr startWhile
    
    while_end:

    ret

;======================================================================================================
;  Says how many free slots the entity array has
;
;  Inputs:
;      -
;  Outputs:
;      - A, number os free space in entity array
;
man_entity_freeSpace::
    ;; Cargo el num_entities y el SIZE_ARRAY (cantidad máx de entities)
    ld ix, #m_num_created_entities

    ld a, #max_entities
    sub a, (ix)   ; Resto al SIZE_ARRAY el num_entities

    ret

;======================================================================================================
;  Updates entity array, deleting if needed
;
;  Inputs:
;      -
;  Outputs:
;      - A, number os free space in entity array
;
man_entity_update::
    ld      ix, #entities      ;; Point before the first entity in order to make the loop easier

    startWhile_update:

        IsValid ix, while_end_update
            ;; Checks if it's dead
            IsDead ix, else_itNotDead
                ;; If it's dead, call man_entity_destroy
                push ix
                call man_entity_destroy
                pop ix

                ;; Do not go to next entity to check the new actual entity
                jr startWhile_update
            
            else_itNotDead:
                ;; If i didn't destroy any entity, go to next
                NextEntity ix   ;; Point to the next entity
                jr startWhile_update
    while_end_update:

    ret


;======================================================================================================
;  Loop through the array of entities and call the funcion passed
;  as a pointer with a pair of entities that matches the signature
;
;  Inputs:
;      - HL, pointer to the function to be called
;      - B, (signature) type of the entity to be called when match the type
;
;  Function's Inputs:
;      - IX, pointer to left entity
;      - IY, pointer to right entity
;
man_entity_forall_pairs_matching::
    ld      ix, #entities-sizeof_e      ;; Point before the first entity in order to make the loop easier

    firstWhile:

        NextEntity ix   ;; entity_left = entities;

        IsValid ix, firstWhile_end
            ;; If the signature does not match, go to the next entity
            SignatureDoesNotMatch ix, b, firstWhile

            ;; If the signature matches, call function pointer
            push    ix  ;; / IY is the right entity, and is going to start at the left entity plus 1,
            pop     iy  ;; \ but this plus one is on the "NextEntity iy" line
            secondWhile:
                NextEntity iy   ;; entity_right = entity_left + 1;
                
                IsValid iy, secondWhile_end
                    SignatureDoesNotMatch iy, b, secondWhile

                    saveRegisters
                    push iy

                    ld      de, #return_all_matching_pairs
                    push    de

                    jp      (hl)    ;; Jump to the function pointer

                    return_all_matching_pairs:
                    pop iy
                    RecoverRegisters

                    jr secondWhile
            secondWhile_end:

            jr firstWhile
    
    firstWhile_end:

    ret

;======================================================================================================
;  Decrement by one the number of lifes of an entity, if it has already zero lifes does not decrement
;
;  Inputs:
;      - IX, pointer to the entity
;
man_entity_lose_life::
    ld a, e_lifes(ix) ; 2
    cp #0
    ret Z

        dec a
        ld e_lifes(ix), a

    ret

man_entity_increase_life::
    ld a, e_lifes(ix) ; 2
    cp #3
    ret Z

        inc a
        ld e_lifes(ix), a

    ret


;======================================================================================================
;  Manages player's collitions 
;
man_entity_decrement_collding_counter::

    ld  ix, #entities

    IsCounterZero ix, _end
        ld  a, e_colliding_counter(ix)
        sub #1
        ld  e_colliding_counter(ix), a

        ld  a, e_colliding_counter(ix)
        cp  #1
        jr C, was_1_and_now_is_zero
            jr _end
        was_1_and_now_is_zero:
            ; Remove invulnerability
            ld  de, #CPCT_VMEM_START_ASM
            ld  a, #0x00
            ld  c, #11 ;; This should be sprites variables but is bugged so i have hardcoded the variables
            ld  b, #7
            call cpct_drawSolidBox_asm
    _end:
    ret
