; MIT License

; Copyright (C) 2022 Z80Enjoyers / Luis Jesús Marhuenda Tendero (@ljmt1) / Alberto Vegara Belmonte (@avb52-ua) / Samuel Benjamín Oliva Bulpitt (@samuel0122)

; Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal 
; in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
; copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

; The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
; FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

;;======================================================================================================
.module AI_System
;;======================================================================================================
.area _CODE
;;======================================================================================================
.include "cpctelera.h.s"
.include "cpct_functions.inc"
.include "man/entity.h.s"
.include "man/levels.h.s"
;;======================================================================================================

.macro CheckJumpIfItsTopBoundAI E, Tag
    ;; VY is negative -> its going up
    ;; Check not crashing to top_bound
    ld  a, e_vy(E)
    add e_y(E)
    cp #SCREEN_TOP_BOUND

    ;; If vy+y < top_bound -> its going up, so reset
    jr C, Tag
.endm

.macro CheckJumpIfItsBottomBoundAI E, Tag
    ;; VY is positive -> its going down
    ;; Check not crashing to bottom_bound
    ld  a, e_vy(E)
    add e_y(E)
    add e_h(E)
    cp #SCREEN_BOTTOM_BOUND

    ;; If vy+y > bottom_bound - h -> its going down, so reset
    jr NC, Tag
.endm


    BEHAVIOUR_UP_DOWN_COUNT     == 30
    BEHAVIOUR_DO_SQUARES_COUNT  == 0x10

;======================================================================================================
;  Executes follow player behaviour
;
;  Inputs:
;      - IX: pointer to entity
;  Outputs:
;      - 
;
sys_ai_behaviour_follow_player::
    ld iy, #entities

    ;; Obtengo las posiciones Y
    ld a, e_y(iy)   ;; playerY
    ld b, e_y(ix)   ;; behaviourY

    cp b
    jr Z, dont_move_y       ;; behaviourY == playerY
    jr C, decrementBehaviourY        ;; behaviourY > playerY

    incrementBehaviourY:
        ld e_vy(ix), #1   
        ret

    dont_move_y:
        ld e_vy(ix), #0 
        ret 

    decrementBehaviourY:
        ld e_vy(ix), #-1   

    ret


;======================================================================================================
;  Executes down-up behaviour
;
;  Inputs:
;      - IX: pointer to entity
;  Outputs:
;      - 
;
sys_ai_behaviour_down_up::

    ld e_behaviour_count(ix), #BEHAVIOUR_UP_DOWN_COUNT

    ;; Go down
    ld e_vy(ix), #1

    ;; Insert normal up-down behaviour
    ld hl, #sys_ai_behaviour_up_down
    ld e_behaviour_func(ix), l
    ld e_behaviour_func+1(ix), h

    ret



;======================================================================================================
;  Executes up-down behaviour
;
;  Inputs:
;      - IX: pointer to entity
;  Outputs:
;      - 
;
sys_ai_behaviour_up_down::
    ;; Compruebo que no me salga
    ld  a, e_vy(ix)
    and #0x80
    jr Z, init_vy_is_positive_behaviour_up_down
        CheckJumpIfItsTopBoundAI ix, resetAndChange_vy_up_down
        jr continue_behaviour_up_down

    init_vy_is_positive_behaviour_up_down:
        CheckJumpIfItsBottomBoundAI ix, resetAndChange_vy_up_down
    
    continue_behaviour_up_down:

    ;; Compruebo el contador
    ld a, e_behaviour_count(ix)
    or a

    ;; Si no es 0, no cambio el behaviour y decremento contador
    jr Z, resetAndChange_vy_up_down
        dec a
        ld e_behaviour_count(ix), a
        ret

    resetAndChange_vy_up_down:

    ;; Si es 0, primero reinicio el contador
    ld e_behaviour_count(ix), #BEHAVIOUR_UP_DOWN_COUNT

    ;; Luego cambio la vy, para ello miro si es negativa
        
    ld a, e_vy(ix)
    and #0x80

    ;; Si es negativa, salto a ponerla positiva
    jr NZ, change_to_positive_speed_up_down
        ld e_vy(ix), #-1
        ret

    change_to_positive_speed_up_down:
    ld e_vy(ix), #1
    
    ret



;======================================================================================================
;  Executes draw squares behaviour
;
;  Inputs:
;      - IX: pointer to entity
;  Outputs:
;      - 
;
sys_ai_behaviour_do_squares::
    ;; The first 2 bits are for status
    ;; 00 -> going left-down
    ;; 01 -> going up
    ;; 10 -> going left-up
    ;; 11 -> going down

    ;; Compruebo que no me salga
    ld  a, e_vy(ix)
    and #0x80
    jr Z, init_vy_is_positive_behaviour_do_squares
        CheckJumpIfItsTopBoundAI ix, do_quares_count_is_z
        jr continue_behaviour_do_squares

    init_vy_is_positive_behaviour_do_squares:
        CheckJumpIfItsBottomBoundAI ix, do_quares_count_is_z
    
    continue_behaviour_do_squares:


    ;; Compruebo el contador
    ld  a, e_behaviour_count(ix)
    and #0x3F   ;; Checks 0x0011 111 bits for counter
    or  a

    ;; Si no es 0, no cambio el behaviour y decremento contador
    jr Z, do_quares_count_is_z
        dec e_behaviour_count(ix)
        ret

    do_quares_count_is_z:

    ;; Checks top 2 bits on counter for status
    ld a, e_behaviour_count(ix)
    and #0xC0
    add #BEHAVIOUR_DO_SQUARES_COUNT+0x40   ;; Goes to next stage & reset counter (32)
    ld e_behaviour_count(ix), a
    
    ;; Checks bit 6
    bit 6, a
    jr NZ, not_going_left_do_squares
        ;; If bit 6 is active, goes to left
        ld e_vy(ix), #0
        ld e_vx(ix), #-1
        
        ret
    not_going_left_do_squares:

    ;; Checks bit 7
    bit 7, a

    jr NZ, not_going_down_do_squares
        ;; If bit 7 is active, goes down
        ld e_vy(ix), #1
        ld e_vx(ix), #0

        ret
    
    not_going_down_do_squares:    
    ;; If bit 7 is not active, goes up
    ld e_vy(ix), #-1
    ld e_vx(ix), #0

    ret



;======================================================================================================
;  Executes draw squares behaviour
;
;  Inputs:
;      - IX: pointer to entity
;  Outputs:
;      - 
;
sys_ai_behaviour_stop_and_run::
    
    ;; Checks if the element is stopped to do flittering
    ld a, e_vx(ix)
    or a
    jr NZ, its_moving_stop_and_jump
        ;; Checks if vy is possitive or negative 
        ld a, e_vy(ix)
        and #0x80

        ;; If vy is positive, jump to set to negative
        jr Z, vy_is_positive_stop_and_run
            ;; If vy is negative, set to positive
            ld  e_vy(ix), #4
            jr its_moving_stop_and_jump
        vy_is_positive_stop_and_run:
            ld  e_vy(ix), #-4
    its_moving_stop_and_jump:

    ;; Compruebo el contador
    ld  a, e_behaviour_count(ix)
    or  a

    ;; Si no es 0, no cambio el behaviour y decremento contador
    jr Z, stop_and_run_count_is_z
        dec e_behaviour_count(ix)
        ret

    stop_and_run_count_is_z:

    ;; Checks X speed
    ld a, e_vx(ix)
    or a

    ;; If it's running on X, jump to stop
    jr NZ, stop_for_a_while_stop_and_run
        ;; If bit 6 is active, goes to left
        ld  e_vx(ix), #-3
        ld  e_vy(ix), #0
        ld  e_behaviour_count(ix), #8   ;; For how long it runs
        ret

    stop_for_a_while_stop_and_run:
    ld  e_behaviour_count(ix), #25      ;; For how long it stays stopped
    ld e_vx(ix), #0
    
    ret



;======================================================================================================
;  Executes up behaviour
;
;  Inputs:
;      - IX: pointer to entity
;  Outputs:
;      - 
;
sys_ai_behaviour_up::
    CheckJumpIfItsTopBoundAI ix, change_to_down

    ld e_vy(ix), #-1
    ret

    change_to_down:
    ld e_vy(ix), #1
    ret


sys_ai_behaviour_straight_forward::

    ;; Don't do anything, just move forward
    ld e_vx(ix), #-1


    ret

sys_ai_behaviour_straight_forward_difficult::

    ;; Don't do anything, just move forward
    ld e_vx(ix), #-2


    ret

;======================================================================================================
;  Executes down behaviour
;
;  Inputs:
;      - IX: pointer to entity
;  Outputs:
;      - 
;
sys_ai_behaviour_down::
    ;; Compruebo que no me salga

    CheckJumpIfItsBottomBoundAI ix, change_to_up
    ld e_vy(ix), #1

    ret

    change_to_up:
    ld e_vy(ix), #-1
    ret


;======================================================================================================
;  Kills the entities after the message counter
;
;  Inputs:
;      - IX: pointer to entity
;  Outputs:
;      - 
;
sys_ai_behaviour_kill_level_message::
    ;; Reduces counter
    dec e_behaviour_count(ix)
    ret NZ

    ;; If counter is 0, kill
    call man_entity_set4destroy
    ret

;======================================================================================================
;  Kills the entities after the message counter
;
;  Inputs:
;      - IX: pointer to entity
;  Outputs:
;      - 
;
sys_ai_behaviour_kill_alarm_message::


    ;; Checks if vy is possitive or negative 
    ld a, e_vx(ix)
    and #0x80

    ;; If vy is positive, jump to set to negative
    jr Z, vy_is_positive_alarm_message
        ;; If vy is negative, set to positive
        ld  e_vx(ix), #1
        jr continue_kill_alarm_message
    vy_is_positive_alarm_message:
        ld  e_vx(ix), #-1
    continue_kill_alarm_message:


    ;; Reduces counter
    dec e_behaviour_count(ix)
    ret NZ

    ;; If counter is 0, kill
    call man_entity_set4destroy
    ret

;======================================================================================================
;  Updates one entity's AI
;
;  Inputs:
;      - IX: pointer to entity
;  Outputs:
;      - 
;
sys_ai_update_one_entity:
    ld l, e_behaviour_func(ix)
    ld h, e_behaviour_func+1(ix)

    jp (hl)


;======================================================================================================
;  System AI update
;
sys_ai_update::

    ld    hl, #sys_ai_update_one_entity
    ld    b, #e_cmp_ai
    call  man_entity_forall_matching

    ret