; MIT License

; Copyright (C) 2022 Z80Enjoyers / Luis Jesús Marhuenda Tendero (@ljmt1) / Alberto Vegara Belmonte (@avb52-ua) / Samuel Benjamín Oliva Bulpitt (@samuel0122)

; Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal 
; in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
; copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

; The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
; FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

;;======================================================================================================
.module Render_System
;;======================================================================================================
.area _CODE
;;======================================================================================================
.include "cpctelera.h.s"
.include "cpct_functions.inc"
.include "assets/sprites.inc"
.include "assets/maps.inc"
.include "man/entity.h.s"
.include "sys/render.h.s"
.include "man/game.h.s"
;;======================================================================================================

currentPalette:
    .db 0

tileCounter:
    .db 0

PALETTE_COUNTER = 5

currentPaletteCounter:
    .db PALETTE_COUNTER

;;======================================================================================================
;;  System Render Init
;;
sys_render_init::
    ; Disables Amstrad CPC firmware, preventing it from being executed at every CPU interrupt.
    call cpct_disableFirmware_asm

    ; (1B C) videoMode	[0-3] Video mode to set
    ld c, #0
    call cpct_setVideoMode_asm

    ; (2B HL) colour_array	Pointer to a byte array containing new hardware colour values [0 - 31]
    ; (2B DE) size	[1 - 16] Number of colours to change
    ld hl, #_global_palette
    ld de, #16
    call cpct_setPalette_asm

    ; (1B) HWC	Hardware colour value for the screen border

    cpctm_setBorder_asm HW_BLACK
    
    ret

;;======================================================================================================
;;  System Render Update One Entity
;;
sys_render_update_one_entity:
    ;; Get prev pointer
    ld  d, e_prev_pointer(ix) 
    ld  e ,e_prev_pointer+1(ix)

    ;; Checks prev pointer != null
    ld  a, d
    or e
    jr Z, dont_clear_prev_pointer
        ;; IF != null -> draw solid box
        ld       a, #0x00    ;; Color
        ld       c, e_w(ix)
        ld       b, e_h(ix)
        call cpct_drawSolidBox_asm

    dont_clear_prev_pointer:
    ld      de, #CPCT_VMEM_START_ASM
    ld      c, e_x(ix)
    ld      b, e_y(ix)
    call cpct_getScreenPtr_asm

    ;; If type is dead don't render
    ld a,    e_type(ix)
    and #e_type_dead
    ret NZ

    ex de, hl

    ld  e_prev_pointer(ix)  , d
    ld  e_prev_pointer+1(ix), e

    ; (2B HL) sprite	Source Sprite Pointer (array with pixel data)
    ; (2B DE) memory	Destination video memory pointer
    ; (1B C ) width	Sprite Width in bytes [1-63] (Beware, not in pixels!)
    ; (1B B ) height	Sprite Height in bytes (>0)
    ld      l, e_sprite+0(ix)
    ld      h, e_sprite+1(ix)
    ld      c, e_w(ix)
    ld      b, e_h(ix)
    call cpct_drawSprite_asm

    ret

;;======================================================================================================
;;  System Render Update
;;
sys_render_update::
    ld    hl, #sys_render_update_one_entity
    ld    b, #e_cmp_render
    call  man_entity_forall_matching

    ret


;;======================================================================================================
;;  System Render Draw win message
;;
sys_render_draw_win_message::
    ld      de, #CPCT_VMEM_START_ASM
    ld      c, #40 - (SPR_WIN_W/2)
    ld      b, #100- (SPR_WIN_H/2)
    call    cpct_getScreenPtr_asm

    ex      de, hl

    ld      hl, #_spr_win_message
    ld      c, #SPR_WIN_W
    ld      b, #SPR_WIN_H
    call cpct_drawSprite_asm
    
    ret

;;======================================================================================================
;;  System Render Draw DEAD message
;;
sys_render_draw_dead_message::
    ld      de, #CPCT_VMEM_START_ASM
    ld      c, #40 - (SPR_DEAD_W/2)
    ld      b, #100- (SPR_DEAD_H/2)
    call    cpct_getScreenPtr_asm

    ex      de, hl

    ld      hl, #_spr_dead_message
    ld      c, #SPR_DEAD_W
    ld      b, #SPR_DEAD_H
    call cpct_drawSprite_asm
    
    ret

;;======================================================================================================
;;  System Render Tilemap
;;
sys_render_tilemap::
    ; (1B C) width	Width in tiles of the view window to be drawn
    ; (1B B) height	Height in tiles of the view window to be drawn
    ; (2B DE) tilemapWidth	Width in tiles of the complete tilemap
    ; (2B HL) tileset	Pointer to the start of the tileset definition (list of 32-byte tiles)
    ld bc, #0x1914
    ld de, #20
    ld hl, #_t_moon_00
    call cpct_etm_setDrawTilemap4x8_ag_asm

    ; (2B HL) memory	Video memory location where to draw the tilemap (character & 4-byte aligned)
    ; (2B DE) tilemap	Pointer to the upper-left tile of the view to be drawn of the tilemap
    ld hl, #CPCT_VMEM_START_ASM
    ld de, #_tm_moon
    call cpct_etm_drawTilemap4x8_ag_asm
    
    ret

;;======================================================================================================
;;  System Render Tilemap Planets
;;
sys_render_tilemap_planets::
    ; (1B C) width	Width in tiles of the view window to be drawn
    ; (1B B) height	Height in tiles of the view window to be drawn
    ; (2B DE) tilemapWidth	Width in tiles of the complete tilemap
    ; (2B HL) tileset	Pointer to the start of the tileset definition (list of 32-byte tiles)
    ld bc, #0x0314
    ld de, #20
    ld hl, #_t_planet_00
    call cpct_etm_setDrawTilemap4x8_ag_asm

    ; (2B HL) memory	Video memory location where to draw the tilemap (character & 4-byte aligned)
    ; (2B DE) tilemap	Pointer to the upper-left tile of the view to be drawn of the tilemap
    ld hl, #CPCT_VMEM_START_ASM + 160
    ld de, #_tm_planet
    call cpct_etm_drawTilemap4x8_ag_asm
    
    ret

;;======================================================================================================
;;  System Render Show Lifes Score
;;
sys_render_show_lifes::

    ld  ix, #entities

    ld  a, e_lifes(ix)
    cp  #3
    jp  z, lifes_equal_to_3

        cp  #2
        jr  z, lifes_equal_to_2

            cp  #1
            jr  z, lifes_equal_to_1

                ld  hl, #_spr_hearts_1
                ld  de, #CPCT_VMEM_START_ASM+38
                ld  c, #SPR_HEARTS_W
                ld  b, #SPR_HEARTS_H
                call cpct_drawSprite_asm

                ld  hl, #_spr_hearts_1
                ld  de, #CPCT_VMEM_START_ASM+44
                ld  c, #SPR_HEARTS_W
                ld  b, #SPR_HEARTS_H
                call cpct_drawSprite_asm

                ld  hl, #_spr_hearts_1
                ld  de, #CPCT_VMEM_START_ASM+50
                ld  c, #SPR_HEARTS_W
                ld  b, #SPR_HEARTS_H
                call cpct_drawSprite_asm

                jr _end

            lifes_equal_to_1:

                ld  hl, #_spr_hearts_1
                ld  de, #CPCT_VMEM_START_ASM+38
                ld  c, #SPR_HEARTS_W
                ld  b, #SPR_HEARTS_H
                call cpct_drawSprite_asm

                ld  hl, #_spr_hearts_1
                ld  de, #CPCT_VMEM_START_ASM+44
                ld  c, #SPR_HEARTS_W
                ld  b, #SPR_HEARTS_H
                call cpct_drawSprite_asm

                ld  hl, #_spr_hearts_0
                ld  de, #CPCT_VMEM_START_ASM+50
                ld  c, #SPR_HEARTS_W
                ld  b, #SPR_HEARTS_H
                call cpct_drawSprite_asm

                jr _end

        lifes_equal_to_2:

            ld  hl, #_spr_hearts_1
            ld  de, #CPCT_VMEM_START_ASM+38
            ld  c, #SPR_HEARTS_W
            ld  b, #SPR_HEARTS_H
            call cpct_drawSprite_asm

            ld  hl, #_spr_hearts_0
            ld  de, #CPCT_VMEM_START_ASM+44
            ld  c, #SPR_HEARTS_W
            ld  b, #SPR_HEARTS_H
            call cpct_drawSprite_asm
            
            ld  hl, #_spr_hearts_0
            ld  de, #CPCT_VMEM_START_ASM+50
            ld  c, #SPR_HEARTS_W
            ld  b, #SPR_HEARTS_H
            call cpct_drawSprite_asm

            jr _end

    lifes_equal_to_3:

        ld  hl, #_spr_hearts_0
        ld  de, #CPCT_VMEM_START_ASM+38
        ld  c, #SPR_HEARTS_W
        ld  b, #SPR_HEARTS_H
        call cpct_drawSprite_asm

        ld  hl, #_spr_hearts_0
        ld  de, #CPCT_VMEM_START_ASM+44
        ld  c, #SPR_HEARTS_W
        ld  b, #SPR_HEARTS_H
        call cpct_drawSprite_asm
        
        ld  hl, #_spr_hearts_0
        ld  de, #CPCT_VMEM_START_ASM+50
        ld  c, #SPR_HEARTS_W
        ld  b, #SPR_HEARTS_H
        call cpct_drawSprite_asm

    _end:
    
    ret

;;======================================================================================================
;;  System Render Show Distance Score
;;      - Distance = XXXX
;;      - Description: The distance score it is printed from the right to the left,
;;      so the first cpct_drawCharM0_asm print the more right number, and so on.
;;
sys_render_show_distance_score::
    ld l, #10                       ;; /
    ld h, #0                        ;;| Set color to distance and lifes scores
    call cpct_setDrawCharM0_asm     ;; \
    
    ld  hl, (distance_score)

    ld  a, l                        ;; / Print first char at the right
    and #0x0F                       ;;
    add #48                         ;;
    ld e, a                         ;;
    ld hl, #CPCT_VMEM_START_ASM+76  ;;
    call cpct_drawCharM0_asm        ;;

    ld  hl, (distance_score)

    ld a, l
    rra
    rra
    rra
    rra
    and #0x0F
    add #48
    ld e, a
    ld hl, #CPCT_VMEM_START_ASM+72
    call cpct_drawCharM0_asm

    ld  hl, (distance_score)

    ld a, h
    and #0x0F
    add #48
    ld e, a
    ld hl, #CPCT_VMEM_START_ASM+68
    call cpct_drawCharM0_asm

    ld  hl, (distance_score)
    
    ld a, h
    rra
    rra
    rra
    rra
    and #0x0F
    add #48
    ld e, a
    ld hl, #CPCT_VMEM_START_ASM+64
    call cpct_drawCharM0_asm

    ret

;======================================================================================================
;  System Render Change Palette
;
sys_render_update_palette::
    ; Comrpuebo si la paleta debe seguir estando gracias a su contador
    ld hl, #currentPaletteCounter
    dec (hl)
    ld a, (hl)
    cp #0
    ret nz

    ld (hl), #PALETTE_COUNTER

    ld hl, #currentPalette
    ld a, (hl)
    cp #0
    jr z, gw

    inc (hl)
    ld hl, #_global_palette_gw
    jr _endifPalette

    gw:
        dec (hl)
        ld hl, #_global_palette

    _endifPalette:

    ld de, #16
    call cpct_setPalette_asm
    ret


;======================================================================================================
;  System Render Show Dead Counter
;
.globl cpct_drawCharM0_asm
sys_render_show_dead_counter::

    ld  hl, (dead_counter)

    ld  a, l                        
    and #0x0F                       
    add #48                         
    ld e, a                         
    ld hl, #CPCT_VMEM_START_ASM+16  
    call cpct_drawCharM0_asm        

    ld  hl, (dead_counter)

    ld a, l
    rra
    rra
    rra
    rra
    and #0x0F
    add #48
    ld e, a
    ld hl, #CPCT_VMEM_START_ASM+12
    call cpct_drawCharM0_asm

    ld  hl, #_spr_dead_counter2
    ld  de, #CPCT_VMEM_START_ASM+20
    ld  c, #SPR_DEAD_COUNTER_W
    ld  b, #SPR_DEAD_COUNTER_H
    call cpct_drawSprite_asm

    ret

sys_render_print_invulnerability::
    ld  hl, #_spr_invulnerabality
    ld  de, #CPCT_VMEM_START_ASM
    ld  c, #SPR_INVULNERABALITY_W
    ld  b, #SPR_INVULNERABALITY_H
    call cpct_drawSprite_asm
    
    ret